@@ -48,22 +48,24 @@
 /* Compare redirection slot information with what we have */
 #define CLUSTER_REDIR_CMP(c) \
     (SLOT_SOCK(c,c->redir_slot)->port != c->redir_port || \
-    strlen(SLOT_SOCK(c,c->redir_slot)->host) != c->redir_host_len || \
-    memcmp(SLOT_SOCK(c,c->redir_slot)->host,c->redir_host,c->redir_host_len))
+    ZSTR_LEN(SLOT_SOCK(c,c->redir_slot)->host) != c->redir_host_len || \
+    memcmp(ZSTR_VAL(SLOT_SOCK(c,c->redir_slot)->host),c->redir_host,c->redir_host_len))
 
 /* Lazy connect logic */
 #define CLUSTER_LAZY_CONNECT(s) \
     if(s->lazy_connect) { \
         s->lazy_connect = 0; \
-        redis_sock_server_open(s, 1 TSRMLS_CC); \
+        redis_sock_server_open(s TSRMLS_CC); \
     }
 
 /* Clear out our "last error" */
-#define CLUSTER_CLEAR_ERROR(c) \
-    if(c->err) efree(c->err); \
-    c->err = NULL; \
-    c->err_len = 0; \
-    c->clusterdown = 0;
+#define CLUSTER_CLEAR_ERROR(c) do { \
+    if (c->err) { \
+        zend_string_release(c->err); \
+        c->err = NULL; \
+    } \
+    c->clusterdown = 0; \
+} while (0)
 
 /* Protected sending of data down the wire to a RedisSock->stream */
 #define CLUSTER_SEND_PAYLOAD(sock, buf, len) \
@@ -216,8 +218,7 @@ typedef struct redisCluster {
     short clusterdown;
 
     /* The last ERROR we encountered */
-    char *err;
-    int err_len;
+    zend_string *err;
 
     /* The slot our command is operating on, as well as it's socket */
     unsigned short cmd_slot;
